/*-
 * Copyright (c) 1982, 1986, 1991, 1993
 *	The Regents of the University of California.  All rights reserved.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 * 3. All advertising materials mentioning features or use of this software
 *    must display the following acknowledgement:
 *	This product includes software developed by the University of
 *	California, Berkeley and its contributors.
 * 4. Neither the name of the University nor the names of its contributors
 *    may be used to endorse or promote products derived from this software
 *    without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE REGENTS AND CONTRIBUTORS ``AS IS'' AND
 * ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
 * IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
 * FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
 * DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS
 * OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION)
 * HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT
 * LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY
 * OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF
 * SUCH DAMAGE.
 *
 */

#include "dbups.h"
#if NDBUPSS > 0

/*
 * Line discipline for "dumb" Uninterruptable Power Supply,
 * which is controlled via RS-232C status lines
 */

#include <sys/types.h>
#include <sys/param.h>
#include <sys/ioctl.h>
#include <sys/ioctl_compat.h>
#include <sys/proc.h>
#include <sys/systm.h>
#include <sys/sysproto.h>
#include <sys/conf.h>
#include <sys/mount.h>
#include <sys/exec.h>
#include <sys/sysent.h>
#include <sys/lkm.h>
#include <a.out.h>
#include <sys/file.h>
#include <sys/errno.h>
#include <sys/reboot.h>

#include <sys/tty.h>
#include <sys/dumbups.h>

struct tty ttypp;

#ifdef notdef
MISC_MOD (dbups);
#else
static int dbups_load __P((struct lkm_table *lkmtp, int cmd));
static int dbups_unload __P((struct lkm_table *lkmtp, int cmd));
int dbups_mod __P((struct lkm_table *lkmtp, int cmd, int ver)); 
static struct lkm_misc _module = {
	LM_MISC,
	LKM_VERSION,
	"dbups_mod"
};
#endif

static int dbupsopen(dev_t dev, struct tty *tp);
static int dbupsclose(struct tty *tp, int flag);
static int dbupsioctl(struct tty *tp, int cmd, caddr_t data, 
	int flag, struct proc *p);

static void ups_shutdown (int howto, void *arg);

struct linesw dbupsline = {
	dbupsopen,	
	dbupsclose,	
	ttread,	
	ttwrite,
	dbupsioctl,	
	ttyinput,	
	ttstart,	
	ttymodem
};


#define UPSFLAG_ACTIVE	0x0001


struct ups {
	int	upsflags;
	int ldisc;
	struct ups_states	upsstates;
	struct tty			*tp;
};


static struct ups ups[NDBUPSS];


static int
dbups_load( struct lkm_table *lkmtp, int cmd)
{
	int			i;
	struct lkm_misc		*args = lkmtp->private.lkm_misc;
	int			err = 0;	/* default = success*/

	switch( cmd) {
	case LKM_E_LOAD:

		/*
		 * Don't load twice! (lkmexists() is exported by kern_lkm.c)
		 */
		if( lkmexists( lkmtp))
			return( EEXIST);

		/* install in linesw[] */
		if ((i = ldisc_register(LDISC_LOAD, &dbupsline)) < 0) {
			err = ENFILE;
			break;
		}
		/* install ups_shutdown routine */
		if (at_shutdown (ups_shutdown, ups, SHUTDOWN_POST_SYNC) > 0) {
			ldisc_deregister(i);
			return ENFILE;
		}

		args->lkm_offset = i;	/* slot in sysent[]*/
		ups->ldisc = i;

		break;		/* Success*/

	case LKM_E_UNLOAD:
		/* current slot...*/
		i = args->lkm_offset;
		ldisc_deregister(i);
		rm_at_shutdown (ups_shutdown, ups);
		break;		/* Success*/

	default:	/* we only understand load/unload*/
		err = EINVAL;
		break;
	}

	return err;
}


int
dbups_mod( lkmtp, cmd, ver)
struct lkm_table	*lkmtp;
int			cmd;
int			ver;
{
	DISPATCH(lkmtp, cmd, ver, dbups_load, dbups_load, lkm_nullcmd);
}


/*
 * Set modem lines
 * Sets all bits set in upss_sm, and reset all bits set in upss_cm
 */


static int setstate (struct tty *tp, struct proc *p, struct ups_statemask *sm)
{
	int i, err=0;

	err = (*cdevsw[major(tp->t_dev)]->d_ioctl)(tp->t_dev, 
		TIOCMGET, (caddr_t)&i, 0, p);
	if (err > 0)
		return err;
	i |= sm->upss_sm;
	i &= ~sm->upss_cm;
	err = (*cdevsw[major(tp->t_dev)]->d_ioctl)(tp->t_dev, 
		TIOCMSET, (caddr_t)&i, 0, p);
	return err;
}


/*
 * Check a state
 * Returns -1 if all bits set in upss_sm are set, and all bits set in 
 * upss_cm are cleared in the modem status.
 */

static int getstate (struct tty *tp, struct proc *p, struct ups_statemask *sm)
{
	int i, err=0;

	err = (*cdevsw[major(tp->t_dev)]->d_ioctl)(tp->t_dev, 
		TIOCMGET, (caddr_t)&i, 0, p);
	if (err > 0)
		return err;
	return ((i & sm->upss_sm) == sm->upss_sm) &&
		!(i & sm->upss_cm) ? -1 : 0;
}


/*
 * Power off the system via the UPS
 */

void ups_shutdown (int howto, void *arg)
{
	struct ups *ups = arg;

	if ((ups->upsflags & UPSFLAG_ACTIVE) && (howto & RB_POWEROFF)) {
		setstate (ups->tp, curproc, &ups->upsstates.ups_off);
	}
}


/*
 */
/*ARGSUSED*/
static int
dbupsopen(dev_t dev, struct tty *tp)
{
	register struct ups *up = ups;	/* XXX only one supported */

	if (tp->t_line == ups->ldisc)
		return (ENODEV);
	ttyflush(tp, FREAD | FWRITE);

	up->upsflags = 0;
	up->tp = tp;

	tp->t_sc = (void *)up;

	return (0);
}

/*
 * Line discipline change or last device close.
 */
static int
dbupsclose (register struct tty *tp, int flag)
{
	return 0;
}


/*ARGSUSED*/
static int
dbupsioctl(struct tty *tp, int cmd, caddr_t data, int flag, struct proc *p)
{
	register struct ups *up = (struct ups *)tp->t_sc;
	int err = 0;

	switch (cmd) {
		case BIOGSTMASK:
			bcopy (&up->upsstates, (void *)data, sizeof (up->upsstates));
			break;

		case BIOSSTMASK:
			up->upsflags = UPSFLAG_ACTIVE;
			bcopy ((void *)data, &up->upsstates, sizeof (up->upsstates));
			setstate (tp, p, &up->upsstates.ups_init);
			break;

		case BIOGSTATUS:
			*(int *)data = 0;
			err = getstate (tp, p, &up->upsstates.ups_linefail);
			if (err > 0)
				return err;
			if (err < 0)
				*(int *)data |= UPSSTATUS_LINEFAIL;
			err = getstate (tp, p, &up->upsstates.ups_battfail);
			if (err > 0)
				return err;
			if (err < 0)
				*(int *)data |= UPSSTATUS_BATTFAIL;
			break;

		default:
			return (-1);		/* pass thru... */

	}
	return 0;
}
#endif
